# -*- coding: utf-8 -*-
# Copyright (c) 2022-2024 Salvador E. Tropea
# Copyright (c) 2022-2024 Instituto Nacional de Tecnología Industrial
# License: AGPL-3.0
# Project: KiBot (formerly KiPlot)
# The Assembly image is a composition from Pixlok and oNline Web Fonts
# The rest are KiCad icons
"""
Dependencies:
  - from: RSVG
    role: Create outputs preview
    id: rsvg1
  - from: RSVG
    role: Create PNG icons
    id: rsvg2
  - from: Ghostscript
    role: Create outputs preview
  - from: ImageMagick
    role: Create outputs preview
  - from: Git
    role: Find origin url
"""
import base64
import os
import subprocess
import pprint
from shutil import copy2
from math import ceil
from struct import unpack
from .bom.kibot_logo import KIBOT_LOGO, KIBOT_LOGO_W, KIBOT_LOGO_H
from .error import KiPlotConfigurationError
from .gs import GS
from .optionable import Optionable, BaseOptions
from .kiplot import config_output, get_output_dir, run_command
from .misc import W_NOTYET, W_MISSTOOL, W_NOOUTPUTS, read_png, force_list
from .pre_base import BasePreFlight
from .registrable import RegOutput
from .macros import macros, document, output_class  # noqa: F401
from . import log, __version__

logger = log.get_logger()
CAT_IMAGE = {'PCB': 'pcbnew',
             'Schematic': 'eeschema',
             'Compress': 'zip',
             'fabrication': 'fabrication',
             'export': 'export',
             'assembly': 'assembly_simple',
             'repair': 'repair',
             'docs': 'project',
             'BoM': 'bom',
             '3D': '3d',
             'gerber': 'gerber',
             'drill': 'load_drill',
             'Auxiliar': 'repair'}
EXT_IMAGE = {'gbr': 'file_gbr',
             'gtl': 'file_gbr',
             'gtp': 'file_gbr',
             'gbo': 'file_gbr',
             'gto': 'file_gbr',
             'gbs': 'file_gbr',
             'gbl': 'file_gbr',
             'gts': 'file_gbr',
             'gml': 'file_gbr',
             'gm1': 'file_gbr',
             'gbrjob': 'file_gerber_job',
             'brd': 'file_brd',
             'bz2': 'file_bz2',
             'dxf': 'file_dxf',
             'cad': 'file_cad',
             'drl': 'file_drl',
             'pdf': 'file_pdf',
             'txt': 'file_txt',
             'pos': 'file_pos',
             'csv': 'file_csv',
             'svg': 'file_svg',
             'eps': 'file_eps',
             'png': 'file_png',
             'jpg': 'file_jpg',
             'plt': 'file_plt',
             'ps': 'file_ps',
             'rar': 'file_rar',
             'scad': 'file_scad',
             'stl': 'file_stl',
             'step': 'file_stp',
             'stp': 'file_stp',
             'wrl': 'file_wrl',
             'html': 'file_html',
             'css': 'file_css',
             'xml': 'file_xml',
             'tsv': 'file_tsv',
             'xlsx': 'file_xlsx',
             'xyrs': 'file_xyrs',
             'xz': 'file_xz',
             'gz': 'file_gz',
             'tar': 'file_tar',
             'zip': 'file_zip',
             'kicad_pcb': 'pcbnew',
             'sch': 'eeschema',
             'kicad_sch': 'eeschema',
             'blend': 'file_blend',
             'pcb3d': 'file_pcb3d',
             'json': 'file_json'}
for i in range(31):
    n = str(i)
    EXT_IMAGE['gl'+n] = 'file_gbr'
    EXT_IMAGE['g'+n] = 'file_gbr'
    EXT_IMAGE['gp'+n] = 'file_gbr'
CAT_REP = {'PCB': ['pdf_pcb_print', 'svg_pcb_print', 'pcb_print'],
           'Schematic': ['pdf_sch_print', 'svg_sch_print']}
BIG_ICON = 256
MID_ICON = 64
OUT_COLS = 12
BIG_2_MID_REL = int(ceil(BIG_ICON/MID_ICON))
IMAGEABLES_SIMPLE = {'png', 'jpg'}
IMAGEABLES_GS = {'pdf', 'eps', 'ps'}
IMAGEABLES_SVG = {'svg'}
TITLE_HEIGHT = 30
STYLE = """
.cat-table { margin-left: auto; margin-right: auto; }
.cat-table td { padding: 20px 24px; }
.nav-table { margin-left: auto; margin-right: auto; }
.nav-table td { padding: 20px 24px; }
.output-table {
  width: 1280px;
  margin-left: auto;
  margin-right: auto;
  border-collapse:
  collapse;
  margin-top: 5px;
  margin-bottom: 4em;
  font-size: 0.9em;
  font-family: sans-serif;
  min-width: 400px;
  border-radius: 5px 5px 0 0;
  overflow: hidden;
  box-shadow: 0 0 20px rgba(0, 0, 0, 0.15);
}
.output-table thead tr { background-color: #0e4e8e; color: #ffffff; text-align: left; }
.output-table th { padding: 10px 12px; }
.output-table td { padding: 5px 7px; }
.out-cell { width: 128px; text-align: center }
.out-img { text-align: center; margin-left: auto; margin-right: auto; }
.cat-img { text-align: center; margin-left: auto; margin-right: auto; }
.td-small { text-align: center; font-size: 0.6em; }
.td-normal { text-align: center; }
.generator { text-align: right; font-size: 0.6em; }
a:link, a:visited { text-decoration: none;}
a:hover, a:active { text-decoration: underline;}
/* The side navigation menu */
.sidenav {
  height: 100%; /* 100% Full-height */
  width: 0; /* 0 width - change this with JavaScript */
  position: fixed; /* Stay in place */
  z-index: 1; /* Stay on top */
  top: 0; /* Stay at the top */
  left: 0;
  background-color: #0e4e8e; /* Black*/
  overflow-x: hidden; /* Disable horizontal scroll */
  padding-top: 60px; /* Place content 60px from the top */
  transition: 0.5s; /* 0.5 second transition effect to slide in the sidenav */
}
/* The navigation menu links */
.sidenav a {
  padding: 8px 8px 8px 8px;
  text-decoration: none;
  font-size: 16px;
  color: #f1f1f1;
  display: block;
  transition: 0.3s;
}
/* When you mouse over the navigation links, change their color */
.sidenav a:hover {
  color: #ff0000;
}
/* Position and style the close button (top right corner) */
.sidenav .closebtn {
  position: absolute;
  top: 0;
  right: 8px;
  font-size: 36px;
  margin-left: 50px;
}
/* Style page content - use this if you want to push the page content to the right when you open the side navigation */
#main {
  transition: margin-left .5s;
  padding: 20px;
  margin-top: @TOP_MAR@px;
}
/* On smaller screens, where height is less than 450px, change the style of the sidenav (less padding and a smaller font
   size) */
@media screen and (max-height: 450px) {
  .sidenav {padding-top: 15px;}
  .sidenav a {font-size: 18px;}
}
ul {
  display: block;
  list-style-type: none;
  margin-block: -1em 0px;
  margin-inline: 0px 0px;
  padding-inline-start: 10px;
}
ul li {
  margin-block: 0px -1em;
}
.topmenu {
  overflow: hidden;
  position: fixed; /* Set the navbar to fixed position */
  top: 0; /* Position the navbar at the top of the page */
  width: 100%; /* Full width */
  background-color: white; /* Otherwise is transparent and overlaps */
}
"""
SCRIPT = """
<script>
function openNav() {
  document.getElementById("theSideNav").style.width = "360px";
  document.getElementById("main").style.marginLeft = "360px";
  document.getElementById("theTopMenu").style.marginLeft = "360px";
  document.getElementById("bmenu").style.display = "none";
}

function closeNav() {
  document.getElementById("theSideNav").style.width = "0";
  document.getElementById("main").style.marginLeft= "0";
  document.getElementById("theTopMenu").style.marginLeft = "0";
  document.getElementById("bmenu").style.display = "block";
}

function ScrollUp() {
  /* When we come here from the navbar we must scroll to avoid the top menu */
  var p = document.getElementById("main");
  var style = p.currentStyle || window.getComputedStyle(p);
  var m_top = parseInt(style.marginTop)

  window.scrollBy(0, -(m_top + 5));
}

window.onload = ScrollUp;
</script>
"""


def _run_command(cmd):
    logger.debug('- Executing: '+GS.pasteable_cmd(cmd))
    try:
        cmd_output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        if e.output:
            logger.debug('Output from command: '+e.output.decode())
        logger.non_critical_error(f'Failed to run {cmd[0]}, error {e.returncode}')
        return False
    if cmd_output.strip():
        logger.debug('- Output from command:\n'+cmd_output.decode())
    return True


def get_png_size(file):
    with open(file, 'rb') as f:
        s = f.read()
    if not (s[:8] == b'\x89PNG\r\n\x1a\n' and (s[12:16] == b'IHDR')):
        return 0, 0
    w, h = unpack('>LL', s[16:24])
    return int(w), int(h)


class Navigate_ResultsOptions(BaseOptions):
    def __init__(self):
        with document:
            self.output = GS.def_global_output
            """ *Filename for the output (%i=html, %x=navigate) """
            self.link_from_root = ''
            """ *The name of a file to create at the main output directory linking to the home page """
            self.skip_not_run = False
            """ Skip outputs with `run_by_default: false` """
            self.logo = Optionable
            """ [string|boolean=''] PNG file to use as logo, use false to remove.
                The KiBot logo is used by default """
            self.logo_url = 'https://github.com/INTI-CMNB/KiBot/'
            """ Target link when clicking the logo """
            self.title = ''
            """ Title for the page, when empty KiBot will try using the schematic or PCB title.
                If they are empty the name of the project, schematic or PCB file is used.
                You can use %X values and KiCad variables here """
            self.title_url = Optionable
            """ [string|boolean=''] Target link when clicking the title, use false to remove.
                KiBot will try with the origin of the current git repo when empty """
            self.nav_bar = True
            """ Add a side navigation bar to quickly access to the outputs """
            self.header = True
            """ Add a header containing information for the project """
        super().__init__()
        self._expand_id = 'navigate'
        self._expand_ext = 'html'

    def config(self, parent):
        super().config(parent)
        # Logo
        if isinstance(self.logo, bool):
            self.logo = '' if self.logo else None
        elif self.logo:
            self.logo = os.path.abspath(self.logo)
            if not os.path.isfile(self.logo):
                raise KiPlotConfigurationError('Missing logo file `{}`'.format(self.logo))
            self._logo_data, self._logo_w, self._logo_h = read_png(self.logo)
            if self._logo_data is None:
                raise KiPlotConfigurationError('Only PNG images are supported for the logo')
        if self.logo == '':
            # Internal logo
            self._logo_w = int(KIBOT_LOGO_W/2)
            self._logo_h = int(KIBOT_LOGO_H/2)
            self._logo_data = base64.b64decode(KIBOT_LOGO)
        elif self.logo is None:
            self._logo_w = self._logo_h = 0
            self._logo_data = ''
        # Title URL
        if isinstance(self.title_url, bool):
            self.title_url = '' if self.title_url else None

    def add_to_tree(self, cat, out, o_tree):
        # Add `out` to `o_tree` in the `cat` category
        cat = cat.split('/')
        node = o_tree
        for c in cat:
            if c not in node:
                # New one
                node[c] = {}
            node = node[c]
        node[out.name] = out

    def svg_to_png(self, svg_file, png_file, width):
        cmd = [self.rsvg_command, '-w', str(width), '-f', 'png', '-o', png_file, svg_file]
        return _run_command(cmd)

    def copy(self, img, width):
        """ Copy an SVG icon to the images/ dir.
            Tries to convert it to PNG. """
        img_w = "{}_{}".format(os.path.basename(img), width)
        if img_w in self.copied_images:
            # Already copied, just return its name
            return self.copied_images[img_w]
        src = os.path.join(self.img_src_dir, img+'.svg') if not img.endswith('.svg') else img
        dst = os.path.join(self.out_dir, 'images', img_w)
        id = img_w
        if self.rsvg_command is not None and self.svg_to_png(src, dst+'.png', width):
            img_w += '.png'
        else:
            copy2(src, dst+'.svg')
            img_w += '.svg'
        name = os.path.join('images', img_w)
        self.copied_images[id] = name
        return name

    def can_be_converted(self, ext):
        if ext in IMAGEABLES_SVG and self.rsvg_command is None:
            logger.warning(W_MISSTOOL+"Missing SVG to PNG converter")
            return False
        if ext in IMAGEABLES_GS and not self.ps2img_avail:
            logger.warning(W_MISSTOOL+"Missing PS/PDF to PNG converter")
            return False
        if ext in IMAGEABLES_SIMPLE and self.convert_command is None:
            logger.warning(W_MISSTOOL+"Missing ImageMagick converter")
            return False
        return ext in IMAGEABLES_SVG or ext in IMAGEABLES_GS or ext in IMAGEABLES_SIMPLE

    def get_image_for_cat(self, cat):
        img = None
        # Check if we have an output that can represent this category
        if cat in CAT_REP and self.convert_command is not None:
            outs_rep = CAT_REP[cat]
            rep_file = None
            # Look in all outputs
            for o in RegOutput.get_outputs():
                # Is this one that can be used to represent it?
                if o.type in outs_rep:
                    out_dir = get_output_dir(o.dir, o, dry=True)
                    targets = o.get_targets(out_dir)
                    # Look the output targets
                    for tg in targets:
                        ext = os.path.splitext(tg)[1][1:].lower()
                        # Can be converted to an image?
                        if os.path.isfile(tg) and self.can_be_converted(ext):
                            rep_file = tg
                            break
                    if rep_file:
                        break
            if rep_file:
                cat, _ = self.get_image_for_file(rep_file, cat, no_icon=True)
                return cat
        if cat in CAT_IMAGE:
            img = self.copy(CAT_IMAGE[cat], BIG_ICON)
            cat_img = '<img src="{}" alt="{}" width="{}" height="{}">'.format(img, cat, BIG_ICON, BIG_ICON)
            cat = ('<table class="cat-img"><tr><td>{}<br>{}</td></tr></table>'.
                   format(cat_img, cat))
        return cat

    def compose_image(self, file, ext, img, out_name, no_icon=False):
        if not os.path.isfile(file):
            logger.warning(W_NOTYET+"{} not yet generated, using an icon".format(os.path.relpath(file)))
            return False, None, None
        if self.convert_command is None:
            return False, None, None
        # Create a unique name using the output name and the generated file name
        bfname = os.path.splitext(os.path.basename(file))[0]
        fname = os.path.join(self.out_dir, 'images', out_name+'_'+bfname+'.png')
        # Full path for the icon image
        icon = os.path.join(self.out_dir, img)
        if ext == 'pdf':
            # Only page 1
            file += '[0]'
        if ext == 'svg':
            tmp_name = GS.tmp_file(suffix='.png')
            logger.debug('Temporal convert: {} -> {}'.format(file, tmp_name))
            if not self.svg_to_png(file, tmp_name, BIG_ICON):
                return False, None, None
            file = tmp_name
        cmd = [self.convert_command, file,
               # Size for the big icons (width)
               '-resize', str(BIG_ICON)+'x']
        if ext == 'ps':
            # ImageMagick 6.9.11 (and also the one in Debian 11) rotates the PS
            cmd.extend(['-rotate', '90'])
        if not no_icon:
            cmd.extend([  # Add the file type icon
                        icon,
                        # At the bottom right
                        '-gravity', 'south-east',
                        # This is a composition, not 2 images
                        '-composite'])
        cmd.append(fname)
        res = _run_command(cmd)
        if ext == 'svg':
            logger.debug('Removing temporal {}'.format(tmp_name))
            os.remove(tmp_name)
        return res, fname, os.path.relpath(fname, start=self.out_dir)

    def get_image_for_file(self, file, out_name, no_icon=False, image=None):
        ext = os.path.splitext(file)[1][1:].lower()
        wide = False
        # Copy the icon for this file extension
        icon_name = 'folder' if os.path.isdir(file) else EXT_IMAGE.get(ext, 'unknown')
        img = self.copy(image or icon_name, MID_ICON)
        # Full name for the file
        file_full = file
        # Just the file, to display it
        file = os.path.basename(file)
        # The icon size
        height = width = MID_ICON
        # Check if this file can be represented by an image
        if self.can_be_converted(ext):
            # Try to compose the image of the file with the icon
            ok, fimg, new_img = self.compose_image(file_full, ext, img, 'cat_'+out_name, no_icon)
            if ok:
                # It was converted, replace the icon by the composited image
                img = new_img
                # Compute its size
                width, height = get_png_size(fimg)
                # We are using the big size
                wide = True
        # Now add the image with its file name as caption
        ext_img = '<img src="{}" alt="{}" width="{}" height="{}">'.format(img, file, width, height)
        file = ('<table class="out-img"><tr><td>{}</td></tr><tr><td class="{}">{}</td></tr></table>'.
                format(ext_img, 'td-normal' if no_icon else 'td-small', out_name if no_icon else file))
        return file, wide

    def add_back_home(self, f, prev):
        if prev is not None:
            prev += '.html'
            f.write('<table class="nav-table">')
            f.write(' <tr>')
            f.write('  <td><a href="{}"><img src="{}" width="{}" height="{}" alt="go back"></a></td>'.
                    format(prev, self.back_img, MID_ICON, MID_ICON))
            f.write('  <td><a href="{}"><img src="{}" width="{}" height="{}" alt="go home"></a></td>'.
                    format(self.home, self.home_img, MID_ICON, MID_ICON))
            f.write(' </tr>')
            f.write('</table>')
        f.write('<p class="generator">Generated by <a href="https://github.com/INTI-CMNB/KiBot/">KiBot</a> v{}</p>\n'.
                format(__version__))
        f.write('</div>\n')
        if self.nav_bar:
            f.write(SCRIPT)

    def write_head(self, f, title):
        f.write('<!DOCTYPE html>\n')
        f.write('<html lang="en">\n')
        f.write('<head>\n')
        f.write(' <title>{}</title>\n'.format(title if title else 'Main page'))
        f.write(' <meta charset="UTF-8">\n')  # UTF-8 encoding for unicode support
        f.write(' <link rel="stylesheet" href="styles.css">\n')
        f.write(' <link rel="icon" href="favicon.ico">\n')
        f.write('</head>\n')
        f.write('<body>\n')
        f.write(self.navbar)
        f.write(self.top_menu)
        f.write('<div id="main">\n')

    def generate_cat_page_for(self, name, node, prev, category):
        logger.debug('- Categories: '+str(node.keys()))
        with open(os.path.join(self.out_dir, name), 'wt') as f:
            self.write_head(f, category)
            name, ext = os.path.splitext(name)
            # Limit to 5 categories by row
            c_cats = len(node)
            rows = ceil(c_cats/5.0)
            by_row = c_cats/rows
            acc = 0
            f.write('<table class="cat-table">\n<tr>\n')
            for cat, content in node.items():
                if not isinstance(content, dict):
                    continue
                if acc >= by_row:
                    # Flush the table and create another
                    acc = 0
                    f.write('</tr>\n</table>\n<table class="cat-table">\n<tr>\n')
                pname = name+'_'+cat+ext
                self.generate_page_for(content, pname, name, category+'/'+cat)
                f.write(' <td><a href="{}">{}</a></td>\n'.format(pname, self.get_image_for_cat(cat)))
                acc += 1
            f.write('</tr>\n</table>\n')
            self.generate_outputs(f, node)
            self.add_back_home(f, prev)
            f.write('</body>\n</html>\n')

    def generate_outputs(self, f, node):
        for oname, out in node.items():
            if isinstance(out, dict):
                continue
            f.write(f'<table id="{oname}" class="output-table">\n')
            out_name = oname.replace(' ', '_')
            oname = oname.replace('_', ' ')
            oname = oname[0].upper()+oname[1:]
            if out.comment:
                oname += ': '+out.comment
            f.write('<thead><tr><th colspan="{}">{}</th></tr></thead>\n'.format(OUT_COLS, oname))
            out_dir = get_output_dir(out.dir, out, dry=True)
            f.write('<tbody><tr>\n')
            targets, icons = out.get_navigate_targets(out_dir)
            c_targets = len(targets)
            # Make the icons a list with same len as targets
            if icons is None:
                icons = [None]*c_targets
            else:
                c_icons = len(icons)
                if c_icons < c_targets:
                    icons.extend([None]*(c_targets-c_icons))
            if len(targets) == 1:
                tg_rel = os.path.relpath(os.path.abspath(targets[0]), start=self.out_dir)
                img, _ = self.get_image_for_file(targets[0], out_name, image=icons[0] if icons else None)
                f.write('<td class="out-cell" colspan="{}"><a href="{}">{}</a></td>\n'.
                        format(OUT_COLS, tg_rel, img))
            else:
                c = 0
                for tg, icon in zip(targets, icons):
                    if c == OUT_COLS:
                        f.write('</tr>\n<tr>\n')
                        c = 0
                    tg_rel = os.path.relpath(os.path.abspath(tg), start=self.out_dir)
                    img, wide = self.get_image_for_file(tg, out_name, image=icon)
                    # Check if we need to break this row
                    span = 1
                    if wide:
                        span = BIG_2_MID_REL
                        remain = OUT_COLS-c
                        if span > remain:
                            f.write('<td class="out-cell" colspan="{}"></td></tr>\n<tr>\n'.format(remain))
                    # Add a new cell
                    f.write('<td class="out-cell" colspan="{}"><a href="{}">{}</a></td>\n'.format(span, tg_rel, img))
                    c = c+span
                if c < OUT_COLS:
                    f.write('<td class="out-cell" colspan="{}"></td>\n'.format(OUT_COLS-c))
            f.write('</tr>\n')
            # This row is just to ensure we have at least 1 cell in each column
            f.write('<tr>\n')
            for _ in range(OUT_COLS):
                f.write('<td></td>\n')
            f.write('</tr>\n')
            f.write('</tbody>\n')
            f.write('</table>\n')

    def generate_end_page_for(self, name, node, prev, category):
        logger.debug('- Outputs: '+str(node.keys()))
        with open(os.path.join(self.out_dir, name), 'wt') as f:
            self.write_head(f, category)
            name, ext = os.path.splitext(name)
            self.generate_outputs(f, node)
            self.add_back_home(f, prev)
            f.write('</body>\n</html>\n')

    def generate_page_for(self, node, name, prev=None, category=''):
        logger.debug('Generating page for '+name)
        if isinstance(list(node.values())[0], dict):
            self.generate_cat_page_for(name, node, prev, category)
        else:
            self.generate_end_page_for(name, node, prev, category)

    def get_targets(self, out_dir):
        # Listing all targets is too complex, we list the most relevant
        # This is good enough to compress the result
        name = self._parent.expand_filename(out_dir, self.output)
        files = [os.path.join(out_dir, 'images'),
                 os.path.join(out_dir, 'styles.css'),
                 os.path.join(out_dir, 'favicon.ico')]
        if self.link_from_root:
            files.append(os.path.join(GS.out_dir, self.link_from_root))
        self.out_dir = out_dir
        self.get_html_names(self.create_tree(), name, files)
        return files

    def get_html_names_cat(self, name, node, prev, category, files):
        files.append(os.path.join(self.out_dir, name))
        name, ext = os.path.splitext(name)
        for cat, content in node.items():
            if not isinstance(content, dict):
                continue
            pname = name+'_'+cat+ext
            self.get_html_names(content, pname, files, name, category+'/'+cat)

    def get_html_names(self, node, name, files, prev=None, category=''):
        if isinstance(list(node.values())[0], dict):
            self.get_html_names_cat(name, node, prev, category, files)
        else:
            files.append(os.path.join(self.out_dir, name))

    def create_tree(self):
        o_tree = {}
        BasePreFlight.configure_all()
        for n in BasePreFlight.get_in_use_names():
            pre = BasePreFlight.get_preflight(n)
            cat = force_list(pre.get_category())
            if not cat:
                continue
            for c in cat:
                self.add_to_tree(c, pre, o_tree)
        for o in RegOutput.get_outputs():
            if not o.run_by_default and self.skip_not_run:
                # Skip outputs that aren't generated in a regular run
                continue
            config_output(o)
            cat = o.category
            if cat is None:
                continue
            for c in cat:
                self.add_to_tree(c, o, o_tree)
        return o_tree

    def generate_navbar_one(self, node, lvl, name, ext):
        """ Recursively create a menu containing all outputs.
            Using ul and li items """
        indent = ' '+' '*lvl
        code = indent+'<ul>\n'
        indent += ' '
        for k, v in node.items():
            if isinstance(v, dict):
                new_name = name+'_'+k
                code += indent+f'<li><a href="{new_name}{ext}">{k}</a></li>\n'
                code += self.generate_navbar_one(v, lvl+1, new_name, ext)
            else:
                code += indent+f'<li><a href="{name}{ext}#{v.name}">{v.name}</a></li>\n'
        code += indent[:-1]+'</ul>\n'
        return code

    def generate_navbar(self, node, name):
        name, ext = os.path.splitext(name)
        code = '<div id="theSideNav" class="sidenav">\n'
        code += '<a href="javascript:void(0)" class="closebtn" onclick="closeNav()">&times;</a>\n'
        code += self.generate_navbar_one(node, 0, name, ext)
        code += '</div>\n'
        return code

    def generate_top_menu(self):
        # Div for the top info
        fsize = f'{TITLE_HEIGHT}px'
        code = '<div id="theTopMenu" class="topmenu">\n'
        code += ' <table style="width:100%">\n'
        code += '  <tr>\n'
        code += '   <td valign="top" align="left">\n'
        if self.nav_bar:
            code += f'    <span id="bmenu" style="font-size:{fsize};cursor:pointer" onclick="openNav()">&#9776;</span>\n'
        code += '   </td>\n'
        code += '   <td>\n'
        if self.logo is not None and self.header:
            img_name = os.path.join('images', 'logo.png')
            if self.logo_url:
                code += f'     <a href="{self.logo_url}">\n'
            code += '     <img src="'+img_name+'" alt="Logo" width="'+str(self._logo_w)+'" height="'+str(self._logo_h)+'">\n'
            if self.logo_url:
                code += '     </a>\n'
        code += '   </td>\n'
        code += '   <td>\n'
        if self.header:
            if self.title_url:
                code += f'     <a href="{self.title_url}">\n'
            code += f'     <span style="font-size:{fsize};">{self._solved_title}</span>\n'
            if self.title_url:
                code += '     </a>\n'
        code += '   </td>\n'
        code += '  </tr>\n'
        code += ' </table>\n'
        code += '</div>\n'
        return code

    def solve_title(self):
        base_title = None
        if GS.sch:
            base_title = GS.sch.get_title()
        if GS.board and not base_title:
            tb = GS.board.GetTitleBlock()
            base_title = tb.GetTitle()
        if not base_title:
            base_title = GS.pro_basename or GS.sch_basename or GS.pcb_basename or 'Unknown'
        text = self.expand_filename_sch(self.title if self.title else '+')
        if text[0] == '+':
            text = base_title+text[1:]
        self._solved_title = text
        # Now the URL
        if self.title_url is not None and not self.title_url:
            # Empty but not None
            self._git_command = self.check_tool('Git')
            if self._git_command:
                res = ''
                try:
                    res = run_command([self._git_command, 'remote', 'get-url', 'origin'], just_raise=True)
                except subprocess.CalledProcessError:
                    pass
                if res:
                    self.title_url = res

    def run(self, name):
        self.out_dir = os.path.dirname(name)
        self.img_src_dir = GS.get_resource_path('images')
        self.img_dst_dir = os.path.join(self.out_dir, 'images')
        os.makedirs(self.img_dst_dir, exist_ok=True)
        self.copied_images = {}
        name = os.path.basename(name)
        # Create a tree with all the outputs
        o_tree = self.create_tree()
        logger.debug('Collected outputs:\n'+pprint.pformat(o_tree))
        if not o_tree:
            logger.warning(W_NOOUTPUTS+'No outputs for navigate results')
            return
        with open(os.path.join(self.out_dir, 'styles.css'), 'wt') as f:
            if not self.header:
                top_margin = 0 if not self.nav_bar else TITLE_HEIGHT
            else:
                top_margin = str(max(self._logo_h, TITLE_HEIGHT))
            f.write(STYLE.replace('@TOP_MAR@', str(top_margin)))
        self.rsvg_command = self.check_tool('rsvg1')
        self.convert_command = self.check_tool('ImageMagick')
        self.ps2img_avail = self.check_tool('Ghostscript')
        # Create the pages
        self.home = name
        self.back_img = self.copy('back', MID_ICON)
        self.home_img = self.copy('home', MID_ICON)
        copy2(os.path.join(self.img_src_dir, 'favicon.ico'), os.path.join(self.out_dir, 'favicon.ico'))
        # Copy the logo image
        if self.logo is not None and self.header:
            with open(os.path.join(self.out_dir, 'images', 'logo.png'), 'wb') as f:
                f.write(self._logo_data)
        self.solve_title()
        self.navbar = self.generate_navbar(o_tree, name) if self.nav_bar else ''
        self.top_menu = self.generate_top_menu() if self.nav_bar or self.header else ''
        self.generate_page_for(o_tree, name)
        # Link it?
        if self.link_from_root:
            redir_file = os.path.join(GS.out_dir, self.link_from_root)
            rel_start = os.path.relpath(os.path.join(self.out_dir, name), start=GS.out_dir)
            logger.debug('Creating redirector: {} -> {}'.format(redir_file, rel_start))
            with open(redir_file, 'wt') as f:
                f.write('<html>\n<head>\n<meta http-equiv="refresh" content="0; {}"/>'.format(rel_start))
                f.write('</head>\n</html>')


@output_class
class Navigate_Results(BaseOutput):  # noqa: F821
    """ Navigate Results
        Generates a web page to navigate the generated outputs """
    def __init__(self):
        super().__init__()
        # Make it low priority so it gets created after all the other outputs
        self.priority = 10
        with document:
            self.options = Navigate_ResultsOptions
            """ *[dict={}] Options for the `navigate_results` output """
        # The help is inherited and already mentions the default priority
        self.fix_priority_help()
        self._any_related = True

    @staticmethod
    def get_conf_examples(name, layers):
        outs = BaseOutput.simple_conf_examples(name, 'Web page to browse the results', 'Browse')  # noqa: F821
        outs[0]['options'] = {'link_from_root': 'index.html', 'skip_not_run': True}
        return outs



# -*- coding: utf-8 -*-
# Copyright (c) 2024 Salvador E. Tropea
# Copyright (c) 2024 Instituto Nacional de Tecnología Industrial
# License: AGPL-3.0
# Project: KiBot (formerly KiPlot)
# Helper functions to draw on a BOARD object
from ..gs import GS
import pcbnew
if GS.ki7:
    # Is this change really needed??!!! People doesn't have much to do ...
    GR_TEXT_HJUSTIFY_LEFT = pcbnew.GR_TEXT_H_ALIGN_LEFT
    GR_TEXT_HJUSTIFY_RIGHT = pcbnew.GR_TEXT_H_ALIGN_RIGHT
    GR_TEXT_HJUSTIFY_CENTER = pcbnew.GR_TEXT_H_ALIGN_CENTER
    GR_TEXT_VJUSTIFY_TOP = pcbnew.GR_TEXT_V_ALIGN_TOP
    GR_TEXT_VJUSTIFY_CENTER = pcbnew.GR_TEXT_V_ALIGN_CENTER
    GR_TEXT_VJUSTIFY_BOTTOM = pcbnew.GR_TEXT_V_ALIGN_BOTTOM
else:
    GR_TEXT_HJUSTIFY_LEFT = pcbnew.GR_TEXT_HJUSTIFY_LEFT
    GR_TEXT_HJUSTIFY_RIGHT = pcbnew.GR_TEXT_HJUSTIFY_RIGHT
    GR_TEXT_HJUSTIFY_CENTER = pcbnew.GR_TEXT_HJUSTIFY_CENTER
    GR_TEXT_VJUSTIFY_TOP = pcbnew.GR_TEXT_VJUSTIFY_TOP
    GR_TEXT_VJUSTIFY_CENTER = pcbnew.GR_TEXT_VJUSTIFY_CENTER
    GR_TEXT_VJUSTIFY_BOTTOM = pcbnew.GR_TEXT_VJUSTIFY_BOTTOM


def draw_rect(g, x, y, w, h, layer, filled=False, line_w=10000):
    if not line_w:
        draw_line(g, x, y, x, y, layer)
        x += w
        y += h
        draw_line(g, x, y, x, y, layer)
        return
    nl = pcbnew.PCB_SHAPE(GS.board)
    nl.SetShape(1)
    if filled:
        nl.SetFilled(True)
    pos = nl.GetStart()
    pos.x = x
    pos.y = y
    nl.SetStart(pos)
    pos = nl.GetEnd()
    pos.x = x+w
    pos.y = y+h
    nl.SetEnd(pos)
    nl.SetLayer(layer)
    nl.SetWidth(line_w)
    g.AddItem(nl)
    GS.board.Add(nl)


def draw_line(g, x1, y1, x2, y2, layer, line_w=10000):
    nl = pcbnew.PCB_SHAPE(GS.board)
    pos = nl.GetStart()
    pos.x = x1
    pos.y = y1
    nl.SetStart(pos)
    pos = nl.GetEnd()
    pos.x = x2
    pos.y = y2
    nl.SetEnd(pos)
    nl.SetLayer(layer)
    nl.SetWidth(line_w)
    g.AddItem(nl)
    GS.board.Add(nl)


def draw_text(g, x, y, text, h, w, layer, bold=False, alignment=GR_TEXT_HJUSTIFY_LEFT):
    nt = pcbnew.PCB_TEXT(GS.board)
    nt.SetText(text)
    nt.SetBold(bold)
    nt.SetTextX(x)
    nt.SetTextY(y+h)
    nt.SetLayer(layer)
    nt.SetTextWidth(w)
    nt.SetTextHeight(h)
    nt.SetHorizJustify(alignment)
    nt.SetVertJustify(GR_TEXT_VJUSTIFY_CENTER)
    g.AddItem(nt)
    GS.board.Add(nt)
    return nt, nt.GetTextBox().GetWidth()


def draw_poly(g, points, layer, filled=False, line_w=10000):
    assert not points or len(points) < 3, "A polygon requires at least 3 points"
    sps = pcbnew.SHAPE_POLY_SET()
    chain = pcbnew.SHAPE_LINE_CHAIN()
    for (x, y) in points:
        chain.Append(x, y)
    chain.SetClosed(True)
    sps.AddOutline(chain)
    ps = pcbnew.PCB_SHAPE(GS.board, pcbnew.SHAPE_T_POLY)
    ps.SetPolyShape(sps)
    ps.SetLayer(layer)
    ps.SetFilled(filled)
    ps.SetWidth(line_w)
    g.AddItem(ps)
    GS.board.Add(ps)


def get_text_width(text, w=10000, bold=False):
    nt = pcbnew.PCB_TEXT(GS.board)
    nt.SetText(text)
    nt.SetBold(bold)
    nt.SetTextWidth(w)
    width = nt.GetTextBox().GetWidth()
    return width



# -*- coding: utf-8 -*-
# Copyright (c) 2024 Nguyen Vincent
# Copyright (c) 2024 Salvador E. Tropea
# Copyright (c) 2024 Instituto Nacional de Tecnología Industrial
# License: AGPL-3.0
# Project: KiBot (formerly KiPlot)
# Contributed by Nguyen Vincent (@nguyen-v)
import os
import csv
from .error import KiPlotConfigurationError
from .gs import GS
from .kicad.pcb_draw_helpers import (draw_rect, draw_line, draw_text, get_text_width,
                                     GR_TEXT_HJUSTIFY_LEFT, GR_TEXT_HJUSTIFY_RIGHT,
                                     GR_TEXT_HJUSTIFY_CENTER)
from .kiplot import load_board, get_output_targets, look_for_output
from .misc import W_NOMATCHGRP
from .optionable import Optionable
from .registrable import RegOutput
from .macros import macros, document, pre_class  # noqa: F401
from . import log
logger = log.get_logger()
ALIGNMENT = {'left': GR_TEXT_HJUSTIFY_LEFT,
             'center': GR_TEXT_HJUSTIFY_CENTER,
             'right': GR_TEXT_HJUSTIFY_RIGHT}
VALID_OUTPUT_TYPES = {'bom', 'kibom', 'position'}


class IncTableOutputOptions(Optionable):
    """ Data for a layer """
    def __init__(self, name=None, parent=None):
        super().__init__()
        self._unknown_is_error = True
        with document:
            self.name = ''
            """ *Name of output """
            self.has_header = True
            """ Plot header on the table """
            self.bold_headers = True
            """ Whether or not the headers should be in bold """
            self.vertical_rule_width = 0.1
            """ Width of vertical rules between columns. Use 0 to eliminate it """
            self.horizontal_rule_width = 0.1
            """ Width of vertical rules between rows (doesn't include header)
                Use 0 to eliminate it """
            self.top_rule_width = 0.4
            """ Width of top rule (above header). Use 0 to eliminate it """
            self.bottom_rule_width = 0.4
            """ Width of bottom rule (bottom of table). Use 0 to eliminate it """
            self.header_rule_width = 0.3
            """ Width of rule below header. Use 0 to eliminate it """
            self.border_width = 0.4
            """ Width of border around the table. Use 0 to eliminate it """
            self.column_spacing = 2
            """ Blank space (in number of characters) between columns """
            self.row_spacing = 2
            """ Space (in number of characters) between rows """
            self.text_alignment = 'left'
            """ [left,center,right] Text alignment in the table """
            self.invert_columns_order = False
            """ Invert column order. Useful when inverting PCB texts in PCB Print """
        if name is not None:
            self.name = name
            self.config(parent)

    def __str__(self):
        v = f'{self.name} ({self.text_alignment}'
        if self.invert_columns_order:
            v += ' inverted'
        if self.has_header:
            v += ' header'
        return v+')'

    def config(self, parent):
        super().config(parent)
        self._text_alignment = ALIGNMENT[self.text_alignment]


class IncludeTableOptions(Optionable):
    """ Include table options """
    def __init__(self):
        with document:
            self.outputs = IncTableOutputOptions
            """ *[list(dict)|list(string)|string=?] List of CSV-generating outputs.
                When empty we include all possible outputs """
            self.enabled = True
            """ Enable the check. This is the replacement for the boolean value """
            self.group_name = 'kibot_table'
            """ Name for the group containing the table. The name of the group
                should be <group_name>_X where X is the output name.
                When the output generates more than one CSV use *kibot_table_out[2]*
                to select the second CSV """
        super().__init__()
        self._unknown_is_error = True

    def config(self, parent):
        super().config(parent)
        if isinstance(self.outputs, type):
            # Nothing specified, look for candidates
            self.outputs = [o.name for o in filter(lambda x: x.type in VALID_OUTPUT_TYPES, RegOutput.get_outputs())]
            logger.debug('- Collected outputs: '+str(self.outputs))
        self._outputs = [IncTableOutputOptions(o, self) if isinstance(o, str) else o for o in self.outputs]


class ITColumns:
    def __init__(self, header='', width=10):
        self.header = header  # Column header name
        self.width = width  # Relative width (default to 10)
        self.data = []  # List to hold data for the column


def update_table_group(g, pos_x, pos_y, width, tlayer, ops, out, csv_file):
    if not os.path.isfile(csv_file):
        raise KiPlotConfigurationError(f'Missing `{csv_file}`, create it first using the `{out.name}` output')
    # Purge all content
    for item in g.GetItems():
        GS.board.Delete(item)
    cols = []

    with open(csv_file) as csvfile:
        reader = csv.reader(csvfile, delimiter=out._obj.get_csv_separator())

        if out.has_header:
            headers = next(reader)
            for header in headers:
                cols.append(ITColumns(header=header))
        else:
            first_row = next(reader)
            for _ in range(len(first_row)):
                cols.append(ITColumns())

            # Add the first row data to the cols
            for i, value in enumerate(first_row):
                cols[i].data.append(value)

        # Add the rest of the CSV rows to the column data
        for row in reader:
            for i, value in enumerate(row):
                if i < len(cols):
                    cols[i].data.append(value)

    if out.invert_columns_order:
        cols.reverse()

    measure_table(cols, out)

    total_char_w = sum(c.width_char for c in cols)
    total_rel_w = sum((c.width for c in cols))  # should be equal to 1

    font_w = int(width/total_char_w) if total_char_w else 0

    xpos_x = int(pos_x + out.column_spacing*font_w/2)
    max_row_data = 0
    for c in cols:
        c.w = int(c.width/total_rel_w*width)
        c.x = xpos_x
        if out._text_alignment == GR_TEXT_HJUSTIFY_LEFT:
            c.xoffset = 0
        if out._text_alignment == GR_TEXT_HJUSTIFY_RIGHT:
            c.xoffset = int(c.w - out.column_spacing*font_w)
        elif out._text_alignment == GR_TEXT_HJUSTIFY_CENTER:
            c.xoffset = int(c.w/2 - out.column_spacing*font_w/2)
        xpos_x += c.w
        max_row_data = max(max_row_data, len(c.data))
    y = pos_y

    row_h = out.row_spacing*font_w

    if out.has_header:
        y += int(row_h)  # Space for top rule + column titles + header rule
        draw_line(g, pos_x, y, pos_x + width, y, tlayer, line_w=GS.from_mm(out.header_rule_width))
        # Draw headers
        for c in cols:
            draw_text(g, c.x + c.xoffset, int(pos_y + 0.5*row_h - font_w), c.header, font_w, font_w,
                      tlayer, bold=out.bold_headers, alignment=out._text_alignment)

    # Draw horizontal rules
    for i in range(max_row_data-1):
        rule_y = int(y + (i+1)*row_h)
        draw_line(g, pos_x, rule_y, pos_x+width, rule_y, tlayer, line_w=GS.from_mm(out.horizontal_rule_width))

    table_h = 0
    for c in cols:
        row_y = int(y + row_h/2)
        for d in c.data:
            draw_text(g, c.x + c.xoffset, int(row_y - font_w), d, font_w, font_w, tlayer, alignment=out._text_alignment)
            row_y += row_h
        table_h = int(max(table_h, row_y-pos_y) - row_h/2)

    # Draw top and bottom rules
    draw_line(g, pos_x, pos_y, pos_x + width, pos_y, tlayer, line_w=GS.from_mm(out.top_rule_width))
    draw_line(g, pos_x, pos_y + table_h, pos_x + width, pos_y + table_h, tlayer, line_w=GS.from_mm(out.bottom_rule_width))

    for n, c in enumerate(cols):
        if n > 0:
            vrule_x = int(c.x - out.column_spacing*font_w/2)
            draw_line(g, vrule_x, pos_y, vrule_x, pos_y + table_h, tlayer, line_w=GS.from_mm(out.vertical_rule_width))

    # Draw rectangle around table
    draw_rect(g, pos_x, pos_y, width, table_h, tlayer, line_w=GS.from_mm(out.border_width))


def measure_table(cols, out):
    col_spacing_width = get_text_width('o')*out.column_spacing

    for c in cols:
        max_data_len = max(get_text_width(d) for d in c.data) if c.data else 0
        max_data_width_char = max(len(d) for d in c.data) if c.data else 0
        c.max_len = max(get_text_width(c.header), max_data_len) + col_spacing_width
        c.width_char = max(len(c.header), max_data_width_char) + out.column_spacing

    tot_len = sum(c.max_len for c in cols)

    # Compute relative widths
    for c in cols:
        c.width = c.max_len/tot_len


def update_table(ops, parent):
    logger.debug('Starting include table preflight')
    load_board()
    csv_files = []
    csv_name = []
    out_to_csv_mapping = {}  # Create a mapping of out variable to its corresponding CSV files

    logger.debug('- Analyzing requested outputs')
    for out in ops._outputs:
        if not out.name:
            raise KiPlotConfigurationError('output entry without a name')
        csv = look_for_output(out.name, '`include table`', parent, VALID_OUTPUT_TYPES) if out.name else None
        if not csv:
            logger.debug(f'  - {out.name} no CSV')
            continue
        out._obj = csv
        targets, _, _ = get_output_targets(out.name, parent)

        # Filter targets to include only CSV files
        csv_targets = [file for file in targets if file.endswith('.csv')]

        for file in csv_targets:
            csv_files.append(file)

        # Append the CSV file names (without path and extension) to csv_name
        for file in csv_targets:
            file_name = os.path.basename(file)  # Get the file name
            name_without_ext = os.path.splitext(file_name)[0]  # Remove the extension
            csv_name.append(name_without_ext)

        # Map the CSV file names to the corresponding out variable
        out_to_csv_mapping[out.name] = (out, csv_targets)
        logger.debug(f'  - {out.name} -> {csv_targets}')

    group_found = False  # Flag to track if any group was found with ops.group_name
    updated = False
    group_prefix = ops.group_name + "_"
    group_prefix_l = len(group_prefix)
    logger.debug('- Scanning board groups')
    for g in GS.board.Groups():
        group_name = g.GetName()
        if not group_name.startswith(group_prefix):
            continue
        group_found = True  # A group with ops.group_name was found
        logger.debug('  - '+group_name)

        # Extract the part after <group_name>_
        group_suffix = group_name[group_prefix_l:]
        index = 0
        if group_suffix[-1] == ']':
            index = int(group_suffix[-2])-1
            group_suffix = group_suffix[:-3]
            logger.debug(f'    - {group_suffix} index: {index}')
        out, csv = out_to_csv_mapping.get(group_suffix)
        if not csv:
            logger.warning(W_NOMATCHGRP+f'No output to handle `{group_name}` found')
            continue
        if index < 0 or index >= len(csv):
            msg = f'index {index+1} is out of range, '+('only one CSV available' if len(csv) == 1 else
                                                        f'must be in the [1,{len(csv)}] range')
            raise KiPlotConfigurationError(msg)
        # We know about it
        x1, y1, x2, y2 = GS.compute_group_boundary(g)
        item = g.GetItems()[0]
        layer = item.GetLayer()
        logger.debug(f'    - Found group @{GS.to_mm(x1)},{GS.to_mm(y1)} mm'
                     f' ({GS.to_mm(x2-x1)}x{GS.to_mm(y2-y1)} mm) layer {layer}'
                     f' with name {g.GetName()}')
        update_table_group(g, x1, y1, x2 - x1, layer, ops, out, csv[index])
        updated = True

    if not group_found:
        logger.warning(W_NOMATCHGRP+f'No `{ops.group_name}*` groups found, skipping `include_table` preflight')

    return updated


@pre_class
class Include_Table(BasePreFlight):  # noqa: F821
    """ Include Table
        Draws a table in the PCB from data in a CSV generated by an output. Needs KiCad 7 or newer.
        To specify the position and size of the drawing you should draw a rectangle in your PCB
        with the width and layer you want. |br|
        Then draw another thing inside the rectangle, select both and create a group
        (right mouse button, then Grouping -> Group). Now edit the group and change its name
        to *kibot_table_X* where X should match the name of the output. Consult the
        `group_name` option for details. |br|
        After running this preflight the rectangle will contain the table with the same name.
        Only the width of the table is important, the height will be adjusted.
        Important: This preflight assumes that a separated KiBot run generated the outputs
        needed for the tables """

    def __init__(self):
        super().__init__()
        self._pcb_related = True
        with document:
            self.include_table = IncludeTableOptions
            """ [boolean|dict=false] Use a boolean for simple cases or fine-tune its behavior """

    def __str__(self):
        v = self.include_table
        if isinstance(v, bool):
            return super().__str__()
        return f'{self.type}: {v.enabled} ({[out.name for out in v._outputs]})'

    def config(self, parent):
        super().config(parent)
        if isinstance(self.include_table, bool):
            self._value = IncludeTableOptions()
            self._value.config(self)
        else:
            self._value = self.include_table

    def apply(self):
        if not GS.ki7:
            raise KiPlotConfigurationError('The `include_table` preflight needs KiCad 7 or newer')
        if update_table(self._value, self):
            GS.save_pcb()
